unit Inv_Main;

{**********************************************************************
 Invert Game
 Copyright (c) 1999 by Christian Maas - All Rights Reserved
                       chmaas@hit.sb.sub.de
                       www.chmaas.handshake.de
 **********************************************************************}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, Grids, StdCtrls, Buttons, Menus;

type
  TInvIndex = 0..8;
type
  TInvArray = array [TInvIndex] of integer;

{ the Move array contains the nine possible moves
  Move [0] corresponds to move 1
  ...
  Move [8] corresponds to move 9
}
const
  Move: array [TInvIndex] of TInvArray =
  {   1  2  3  4  5  6  7  8  9   <- representing tiles }
    ((1, 1, 0, 1, 1, 0, 0, 0, 0), { move 1 inverts the four tiles 1, 2, 4, 5 }
     (1, 1, 1, 0, 0, 0, 0, 0, 0), { move 2 inverts the tiles 1, 2, 3 }
     (0, 1, 1, 0, 1, 1, 0, 0, 0), { move 3 }
     (1, 0, 0, 1, 0, 0, 1, 0, 0), { move 4 }
     (1, 1, 1, 1, 1, 1, 1, 1, 1), { move 5 inverts all tiles }
     (0, 0, 1, 0, 0, 1, 0, 0, 1), { move6 }
     (0, 0, 0, 1, 1, 0, 1, 1, 0), { move 7 }
     (0, 0, 0, 0, 0, 0, 1, 1, 1), { move 8 }
     (0, 0, 0, 0, 1, 1, 0, 1, 1));{ move 9 }

{ the MoveRule array contains the rules for inverting each tile
  MoveRule [0] desribes how to invert tile 1
  ...
  MoveRule [8] desribes how to invert tile 9
}
const
  MoveRule: array [TInvIndex] of TInvArray =
    { 1  2  3  4  5  6  7  8  9   <- representing moves }
    ((0, 1, 0, 1, 1, 0, 0, 0, 1), { rule for inverting tile 1 by moves 2, 4, 5, 9 }
     (0, 1, 0, 1, 0, 1, 1, 0, 1), { rule for inverting tile 2 by moves 2, 4, 6, 7, 9 }
     (0, 1, 0, 0, 1, 1, 1, 0, 0), { rule for inverting tile 3 }
     (0, 1, 1, 1, 0, 0, 0, 1, 1), { rule for inverting tile 4 }
     (1, 1, 1, 1, 1, 1, 1, 1, 1), { rule for inverting tile 5 by all moves 1-9       }
     (1, 1, 0, 0, 0, 1, 1, 1, 0), { rule for inverting tile 6 }
     (0, 0, 1, 1, 1, 0, 0, 1, 0), { rule for inverting tile 7 }
     (1, 0, 1, 1, 0, 1, 0, 1, 0), { rule for inverting tile 8 }
     (1, 0, 0, 0, 1, 1, 0, 1, 0));{ rule for inverting tile 9 }

type
  TfrmInvert = class(TForm)
    Panel1: TPanel;
    pnlStatus: TPanel;
    Panel3: TPanel;
    StringGrid1: TStringGrid;
    Timer1: TTimer;
    sbtNew: TSpeedButton;
    sbtMinimalMoves: TSpeedButton;
    MainMenu1: TMainMenu;
    Game1: TMenuItem;
    Help1: TMenuItem;
    New1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    Contents1: TMenuItem;
    About1: TMenuItem;
    MinimalMoves1: TMenuItem;
    sbtAbout: TSpeedButton;
    sbtHelp: TSpeedButton;
    sbtExit: TSpeedButton;
    procedure StringGrid1DrawCell(Sender: TObject; Col, Row: Longint;
      Rect: TRect; State: TGridDrawState);
    procedure Timer1Timer(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure StringGrid1Click(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure New1Click(Sender: TObject);
    procedure MinimalMoves1Click(Sender: TObject);
    procedure Contents1Click(Sender: TObject);
  private
    { Private declarations }
    { current state of tiles:
      TileStatus [0] corresponds to tile 1
      ...
      TileStatus [8] corresponds to tile 9

      TileStatus [i] = 0 means corresponding tile i + 1 is white
      TileStatus [i] = 1 means corresponding tile i + 1 is whiteblue
    }
    TileStatus: TInvArray;
    StartTime: TDateTime; { start time of current game }
    nMove: integer; { number of moves performed during current game }
    nMinMove: integer; { minimal number of moves at begin of current game }
    procedure WMGetMinMaxInfo(var Msg: TWMGetMinMaxInfo); message WM_GETMINMAXINFO;
    procedure WMNCHitTest(var Msg: TWMNCHitTest); message WM_NCHitTest;
    procedure NewGame;
    function CountOf1 (var A: TInvArray): integer;
    procedure RandomInit (var A: TInvArray);
    function ColRowIndex (Col, Row: integer): integer;
    function AddVectorModulo (A, B: TInvArray): TInvArray;
    function MinimalMoves: integer;
    procedure DoMove (i: integer);
  public
    { Public declarations }
  end;

var
  frmInvert: TfrmInvert;

implementation
uses Inv_Tip, Inv_Abt;

{$R *.DFM}

{**********************************************************************
 allow only minimizing of window
 **********************************************************************}

procedure TfrmInvert.WMGetMinMaxInfo (var Msg: TWMGetMinMaxInfo);
begin
  inherited;
  with Msg.MinMaxInfo^ do begin
    { minimizing is allowed , i.e.
      ptMinTrackSize.x := 0;
      ptMinTrackSize.y := 0;
    }
    { current width and height are maximum values,
      i.e. don't allow maximizing }
    ptMaxTrackSize.x := Width;
    ptMaxTrackSize.y := Height;
  end;
end;

{**********************************************************************
 don't show resizing cursor
 **********************************************************************}

procedure TfrmInvert.WMNCHitTest (var Msg: TWMNCHitTest);
begin
  inherited;
  with Msg do
    if Result in [HTLEFT, HTRIGHT, HTBOTTOM, HTBOTTOMRIGHT,
      HTBOTTOMLEFT, HTTOP, HTTOPRIGHT, HTTOPLEFT]
    then
      Result:= HTNOWHERE;
end;

{**********************************************************************
 initialize new game
 **********************************************************************}

procedure TfrmInvert.NewGame;
begin

  { initialize tiles with random values }
  repeat
    RandomInit (TileStatus);
  until CountOf1 (TileStatus) < 9; { not all tiles are 1, i.e. game is not over }

  { repaint grid }
  StringGrid1.Invalidate;

  { enable menu and speed button to display minimal moves }
  MinimalMoves1.Enabled := true;
  sbtMinimalMoves.Enabled := true;

  { determine number of minimal moves (and write moves to memo in form frmTip,
    which is not used here) }
  nMinMove := MinimalMoves;

  nMove := 0;
  StartTime := Time;

  Timer1.Enabled := true;
  { force displaying of nMove and elapsed time }
  Timer1Timer (nil);

end;

{**********************************************************************
 compute tile index from column and row of grid
 Col = 0, Row = 0 corresponds to Index 0
 ...
 Col = 2, Row = 2 corresponds to Index 8
 **********************************************************************}

function TfrmInvert.ColRowIndex (Col, Row: integer): integer;
begin
  Result := Row * 3 + Col;
end;

{**********************************************************************
 determine count of blue tiles (with value 1) in vector A
 **********************************************************************}

function TfrmInvert.CountOf1 (var A: TInvArray): integer;
var i: TInvIndex;
begin
  Result := 0;
  for i := Low (TInvIndex) to High (TInvIndex) do
    inc (Result, A [i]);
end;

{**********************************************************************
 initialize vector A with random values (0, 1)
 **********************************************************************}

procedure TfrmInvert.RandomInit (var A: TInvArray);
var i: TInvIndex;
begin
  for i := Low (TInvIndex) to High (TInvIndex) do
    A [i] := Random (2);
end;

{**********************************************************************
 add the vectors a, b using modulo 2 operation
 **********************************************************************}

function TfrmInvert.AddVectorModulo (A, B: TInvArray): TInvArray;
var i: TInvIndex;
begin
  for i := Low (TInvIndex) to High (TInvIndex) do
    Result [i] := (A [i] + B [i]) mod 2;
end;

{**********************************************************************
 determine optimal moves
 **********************************************************************}

function TfrmInvert.MinimalMoves: integer;
var MoveVector: TInvArray;
    i: TInvIndex;
begin

  { no moves are done }
  for i := Low (TInvIndex) to High (TInvIndex) do
    MoveVector [i] := 0;

  { for each tile with state 0 add move corresponding MoveRule vector
    to MoveVector }
  for i := Low (TInvIndex) to High (TInvIndex) do
    if TileStatus [i] = 0 then
      { perform corresponding move }
      MoveVector := AddVectorModulo (MoveVector, MoveRule [i]);

  Result := CountOf1 (MoveVector);

  with frmTip do begin
    mmoOpt.Lines.Clear;
    for i := Low (TInvIndex) to High (TInvIndex) do
      if MoveVector [i] = 1 then
        { move i + 1 must be performed }
        mmoOpt.Lines.Add ('Move ' + IntToStr (i + 1));
  end;

end;

{**********************************************************************
 do move i + 1, corresponding to Move [i] vector
 **********************************************************************}

procedure TfrmInvert.DoMove (i: integer);
var Msg: array [0..255] of char;
var Msg2: string;
begin
  if Timer1.Enabled then begin
    { add Move [i] to TileStatus }
    TileStatus := AddVectorModulo (TileStatus, Move [i]);
    inc (nMove);
    Timer1Timer (nil);
    StringGrid1.Invalidate;

    if CountOf1 (TileStatus) = 9 then begin
      { game over }
      Timer1.Enabled := false;
      { disable displaying of optimal moves }
      MinimalMoves1.Enabled := false;
      sbtMinimalMoves.Enabled := false;

      if nMove = nMinMove then
        StrFmt (Msg, 'Congratulations!' + #13 +
            'Game over after the minimum of %d move(s).', [nMinMove])
      else
        StrFmt (Msg, 'You need more practice!' + #13 +
          'Game over after %d moves.' + #13 +
          'Minimum number of moves was %d.', [nMove, nMinMove]);

      Application.MessageBox (Msg, 'Game over', MB_ICONINFORMATION or MB_OK);

    end;
  end;
end;

{**********************************************************************
 draw tile:
 - background blue or white, depending on TileStatus
 - number of tile as centered text
 **********************************************************************}

procedure TfrmInvert.StringGrid1DrawCell(Sender: TObject; Col, Row: Longint;
  Rect: TRect; State: TGridDrawState);
var clBack, clText: TColor;
    xPos, yPos: integer;
begin

  if TileStatus [ColRowIndex (Col, Row)] = 0 then begin
    { white tile }
    clBack := clWhite;
    clText := clNavy;
  end
  else begin
    { blue tile }
    clBack := clNavy;
    clText := clWhite;
  end;

  with StringGrid1.Canvas do begin
    { draw background }
    Brush.Color := clBack;
    FillRect (Rect);

    { draw tile number }
    Font.Color := clText;
    { positions to center text }
    xPos := (Rect.Left + Rect.Right - TextWidth ('X')) div 2;
    yPos := (Rect.Top  + Rect.Bottom - TextHeight ('X')) div 2;
    TextOut (xPos, yPos, StringGrid1.Cells [Col, Row]);
  end;

end;

{**********************************************************************
 display current number of moves and elapsed time
 **********************************************************************}

procedure TfrmInvert.Timer1Timer(Sender: TObject);
begin
  pnlStatus.Caption := 'Moves done: ' + IntToStr (nMove) + ', elapsed time: ' +
    FormatDateTime ('nn:ss', Time - StartTime);
end;

{**********************************************************************
 use keyboard for doing move
 **********************************************************************}

procedure TfrmInvert.FormKeyPress(Sender: TObject; var Key: Char);
begin
  if Key in ['1'..'9'] then
    { key i performs move i - 1 }
    DoMove (ord (Key) - ord ('0') - 1);
end;

{**********************************************************************
 use mouse for doing move
 **********************************************************************}

procedure TfrmInvert.StringGrid1Click(Sender: TObject);
var p: TPoint;
    Col, Row: integer;
begin
  { get absolute coordinates }
  GetCursorPos (p);
  { determine relative coordinates within grid }
  p := StringGrid1.ScreenToClient (p);
  { determine Col and Row of grid cell which was clicked }
  Col := p.x div StringGrid1.DefaultColWidth;
  Row := p.y div StringGrid1.DefaultRowHeight;
  DoMove (ColRowIndex (Col, Row));
end;

{**********************************************************************
 - adjust grid width and height
 - initialize grid cells
 **********************************************************************}

procedure TfrmInvert.FormShow(Sender: TObject);
var i, j: integer;
begin

  { initialize generator for random numbers } 
  Randomize;

  StringGrid1.Width := StringGrid1.GridWidth;
  StringGrid1.Height := StringGrid1.GridHeight;

  { center grid }
  StringGrid1.Left := (ClientWidth - StringGrid1.Width) div 2;

  with StringGrid1 do begin
    Width := ColCount * DefaultColWidth + (2 * ColCount - 1) * GridLineWidth;
    Height := RowCount * DefaultRowHeight + (2 * RowCount - 1) * GridLineWidth;
    for i := 0 to RowCount - 1 do
      for j := 0 to ColCount - 1 do
        Cells [j, i] := IntToStr (ColRowIndex (j, i) + 1); (*  i * 3 + j + 1);*)
  end;

  NewGame;

end;

{**********************************************************************
 **********************************************************************}

procedure TfrmInvert.Exit1Click(Sender: TObject);
begin
  close;
end;

{**********************************************************************
 **********************************************************************}

procedure TfrmInvert.About1Click(Sender: TObject);
begin
  frmAbout.ShowModal;
end;

{**********************************************************************
 **********************************************************************}

procedure TfrmInvert.New1Click(Sender: TObject);
begin
  NewGame;
end;

{**********************************************************************
 display minimal moves to solve *CURRENT* state of tiles
 **********************************************************************}

procedure TfrmInvert.MinimalMoves1Click(Sender: TObject);
begin
  MinimalMoves; { function return value (number of moves) ignored }
  frmTip.ShowModal;
end;

{**********************************************************************
 **********************************************************************}

procedure TfrmInvert.Contents1Click(Sender: TObject);
begin
  Application.HelpCommand (HELP_CONTENTS, 0);
end;

end.
